from cfn_sphere.exceptions import NoConfigException
from yaml.scanner import ScannerError
import yaml
import os





class Config(object):
    def __init__(self, config_file=None, config_dict=None):

        if config_dict:
            self.dict = config_dict
            self.working_dir = None
        else:
            self.dict = self._read_config_file(config_file)
            self.working_dir = os.path.dirname(os.path.realpath(config_file))

        if not isinstance(self.dict, dict):
            raise NoConfigException("Config has invalid content, must be of type dict/yaml")

        self.region = self.dict.get('region')
        self.stacks = self._parse_stack_configs(self.dict)

        self._validate()

    def _validate(self):
        try:
            assert self.region, "Please specify region in config file"
            assert isinstance(self.region, str), "Region must be of type str, not {0}".format(type(self.region))
            assert self.stacks, "Please specify stacks in config file"
        except AssertionError as e:
            raise NoConfigException(e)

    def _parse_stack_configs(self, config_dict):
        stacks_dict = {}
        for key, value in config_dict.get('stacks', {}).items():
            stacks_dict[key] = StackConfig(value, self.working_dir)
        return stacks_dict

    @staticmethod
    def _read_config_file(config_file):
        try:
            with open(config_file, 'r') as f:
                return yaml.load(f.read())
        except IOError as e:
            raise NoConfigException("Could not read yaml file: {0}".format(e))
        except ScannerError as e:
            raise NoConfigException("Could not parse yaml {0}: {1}".format(e.problem_mark, e.problem))

    def get(self):
        return self.dict


class StackConfig(object):
    def __init__(self, stack_config_dict, working_dir=None):
        self.parameters = stack_config_dict.get('parameters', {})
        self.timeout = stack_config_dict.get('timeout', 600)
        self.working_dir = working_dir

        try:
            self.template_url = stack_config_dict['template-url']
        except KeyError as e:
            raise NoConfigException("Stack config needs a {0} key".format(e))


if __name__ == "__main__":
    reader = Config("resources/myapp.yml")
    print(reader.get())



from cfn_sphere.exceptions import TemplateErrorException


class CloudFormationTemplateTransformer(object):
    @classmethod
    def transform_template(cls, template):
        template_dict = template.body_dict

        template_dict = cls.transform_dict_values(template_dict, cls.transform_reference_string)
        template_dict = cls.transform_dict_values(template_dict, cls.transform_getattr_string)
        template_dict = cls.transform_dict_keys(template_dict, cls.transform_join_key)
        template_dict = cls.transform_dict_keys(template_dict, cls.transform_taupage_user_data_key)

        template_dict = cls.transform_dict_values(template_dict, cls.check_for_leftover_reference_values)
        template_dict = cls.transform_dict_keys(template_dict, cls.check_for_leftover_reference_keys)

        template.body_dict = template_dict
        return template

    @staticmethod
    def check_for_leftover_reference_values(value):
        if isinstance(value, list):
            for item in value:
                if item.startswith('|'):
                    raise TemplateErrorException("Unhandled reference value found: {0}".format(value))
        elif isinstance(value, basestring):
            if value.startswith('|'):
                raise TemplateErrorException("Unhandled reference value found: {0}".format(value))

        return value

    @staticmethod
    def check_for_leftover_reference_keys(key, value):
        if key.startswith('|'):
            raise TemplateErrorException("Unhandled reference key found: {0}".format(key))
        if key.startswith('@') and key.endswith('@'):
            raise TemplateErrorException("Unhandled reference key found: {0}".format(key))

        return key, value

    @classmethod
    def transform_taupage_user_data_key(cls, key, value):
        if not value:
            return key, value

        if isinstance(key, basestring):

            if str(key).lower() == '@taupageuserdata@':

                if not isinstance(value, dict):
                    raise TemplateErrorException("Value of 'TaupageUserData' must be of type dict")

                lines = ['#taupage-ami-config']
                lines.extend(cls.transform_userdata_dict_to_lines_list(value))

                return "UserData", {
                    'Fn::Base64': {
                        'Fn::Join': ['\n', lines]
                    }
                }

        return key, value

    @classmethod
    def transform_join_key(cls, key, value):
        if not value:
            return key, value

        if isinstance(key, basestring):
            if key.lower().startswith('|join|'):
                if not isinstance(value, list):
                    raise TemplateErrorException("Value of '|join|' must be of type list")

                join_string = key[6:]

                return 'Fn::Join', [join_string, value]

        return key, value

    @staticmethod
    def transform_kv_to_cfn_join(key, value):
        if isinstance(value, basestring) and ':' in key:
            key = "'{0}'".format(key)

        if isinstance(value, basestring) and ':' in value:
            value = "'{0}'".format(value)

        return {'Fn::Join': [': ', [key, value]]}

    @staticmethod
    def transform_reference_string(value):
        if not value:
            return value

        if isinstance(value, basestring):
            if value.lower().startswith('|ref|'):
                referenced_value = value[5:]

                if not referenced_value:
                    raise TemplateErrorException("Reference must be like |ref|resource")

                return {'Ref': referenced_value}

        return value

    @staticmethod
    def transform_getattr_string(value):
        if not value:
            return value

        if isinstance(value, basestring):
            if value.lower().startswith('|getatt|'):
                elements = value.split('|', 3)

                if len(elements) != 4:
                    raise TemplateErrorException("Attribute reference must be like '|getatt|resource|attribute'")

                resource = elements[2]
                attribute = elements[3]

                return {'Fn::GetAtt': [resource, attribute]}

        return value

    @classmethod
    def transform_userdata_dict_to_lines_list(cls, userdata_dict, indentation_level=0):
        lines = []

        for key, value in userdata_dict.items():

            # key indentation with two spaces
            if indentation_level > 0:
                indented_key = '  ' * indentation_level + str(key)
            else:
                indented_key = key

            if isinstance(key, basestring):

                # do not go any further and directly return cfn functions and their values
                if key.lower() == 'ref' or key.lower() == 'fn::getatt' or key.lower() == 'fn::join':
                    return {key: value}
                else:

                    # recursion for dict values
                    if isinstance(value, dict):

                        result = cls.transform_userdata_dict_to_lines_list(value, indentation_level + 1)

                        if isinstance(result, dict):
                            lines.append(cls.transform_kv_to_cfn_join(indented_key, result))
                        elif isinstance(result, list):
                            result.reverse()
                            lines.extend(result)
                            lines.append(indented_key + ':')
                        else:
                            raise TemplateErrorException("Failed to convert user-data dict to list of lines")

                    else:
                        lines.append(cls.transform_kv_to_cfn_join(indented_key, value))
            else:
                lines.append(cls.transform_kv_to_cfn_join(indented_key, value))

        lines.reverse()
        return lines

    @classmethod
    def transform_dict_keys(cls, dictionary, key_handler):
        for key in dictionary:
            value = dictionary[key]

            if isinstance(value, dict):
                dictionary[key] = cls.transform_dict_keys(value, key_handler)

            if isinstance(value, list):
                for item in value:
                    if isinstance(item, dict):
                        cls.transform_dict_keys(item, key_handler)

            new_key, new_value = key_handler(key, value)

            if new_key != key or new_value != value:
                dictionary[new_key] = new_value

            if new_key != key:
                dictionary.pop(key)

        return dictionary

    @classmethod
    def transform_dict_values(cls, dictionary, value_handler):
        for key in dictionary:
            value = dictionary[key]

            if isinstance(value, dict):
                dictionary[key] = cls.transform_dict_values(value, value_handler)

            elif isinstance(value, list):
                value_list = []
                for item in value:
                    if isinstance(item, dict):
                        value_list.append(cls.transform_dict_values(item, value_handler))
                    elif isinstance(item, basestring):
                        value_list.append(value_handler(item))
                    else:
                        value_list.append(item)

                dictionary[key] = value_list

            elif isinstance(value, basestring):
                dictionary[key] = value_handler(value)
            else:
                dictionary[key] = value

        return dictionary


if __name__ == "__main__":
    template = CloudFormationTemplateTransformer



import logging
import time
from datetime import timedelta
from boto import cloudformation
from boto.exception import BotoServerError
from cfn_sphere.util import get_logger, get_message_from_boto_server_error, get_cfn_api_server_time, \
    get_pretty_parameters_string
from cfn_sphere.aws.cloudformation.stack import CloudFormationStack
from cfn_sphere.exceptions import CfnStackActionFailedException


class CloudFormation(object):
    def __init__(self, region="eu-west-1"):
        logging.getLogger('boto').setLevel(logging.FATAL)
        self.logger = get_logger()

        self.conn = cloudformation.connect_to_region(region)
        if not self.conn:
            self.logger.error("Could not connect to cloudformation API in {0}. Invalid region?".format(region))
            raise Exception("Got None connection object")

        self.logger.debug("Connected to cloudformation API at {0} with access key id: {1}".format(
            region, self.conn.aws_access_key_id))

    def get_stacks(self):
        result = []
        response = self.conn.describe_stacks()
        result.extend(response)
        while response.next_token:
            response = self.conn.describe_stacks(next_token=response.next_token)
            result.extend(response)
        return result

    def get_stack_names(self):
        return [stack.stack_name for stack in self.get_stacks()]

    def get_stacks_dict(self):
        stacks_dict = {}
        for stack in self.get_stacks():
            stacks_dict[stack.stack_name] = {"parameters": stack.parameters, "outputs": stack.outputs}
        return stacks_dict

    def get_stack(self, stack_name):
        return self.conn.describe_stacks(stack_name)[0]

    def validate_stack_is_ready_for_updates(self, stack_name):
        stack = self.get_stack(stack_name)
        valid_states = ["CREATE_COMPLETE", "UPDATE_COMPLETE", "ROLLBACK_COMPLETE", "UPDATE_ROLLBACK_COMPLETE"]

        if stack.stack_status not in valid_states:
            raise CfnStackActionFailedException(
                "Stack {0} is in '{1}' state.".format(stack.stack_name, stack.stack_status))

    def get_stack_state(self, stack_name):
        stack = self.conn.describe_stacks(stack_name)
        return stack.status

    def get_stack_parameters_dict(self, stack_name):
        parameters = {}
        stack = self.get_stack(stack_name)

        for parameter in stack.parameters:
            parameters[parameter.key] = parameter.value

        return parameters

    def create_stack(self, stack):
        assert isinstance(stack, CloudFormationStack)
        try:
            self.logger.info(
                "Creating stack {0} from template {1} with parameters:\n{2}".format(stack.name, stack.template.name,
                                                                                    get_pretty_parameters_string(
                                                                                        stack.parameters)))
            self.conn.create_stack(stack.name,
                                   template_body=stack.template.get_template_json(),
                                   parameters=stack.get_parameters_list(),
                                   capabilities=['CAPABILITY_IAM'])

            self.wait_for_stack_action_to_complete(stack.name, "create", stack.timeout)
            self.logger.info("Create completed for {0}".format(stack.name))
        except BotoServerError as e:
            message = get_message_from_boto_server_error(e)
            raise CfnStackActionFailedException(
                "Could not create stack {0}. Cfn API responded with: {1}".format(stack.name, message))

    def update_stack(self, stack):
        assert isinstance(stack, CloudFormationStack)
        try:
            self.logger.info(
                "Updating stack {0} from template {1} with parameters:\n{2}".format(stack.name, stack.template.name,
                                                                                    get_pretty_parameters_string(
                                                                                        stack.parameters)))

            self.conn.update_stack(stack.name,
                                   template_body=stack.template.get_template_json(),
                                   parameters=stack.get_parameters_list(),
                                   capabilities=['CAPABILITY_IAM'])

            self.wait_for_stack_action_to_complete(stack.name, "update", stack.timeout)
            self.logger.info("Update completed for {0}".format(stack.name))
        except BotoServerError as e:
            message = get_message_from_boto_server_error(e)

            if message == "No updates are to be performed.":
                self.logger.info("Nothing to do: {0}.".format(message))
            else:
                raise CfnStackActionFailedException("{0}.".format(message))

    def wait_for_stack_events(self, stack_name, expected_event, valid_from_timestamp, timeout):

        logging.debug("Waiting for {0} events, newer than {1}".format(expected_event, valid_from_timestamp))

        seen_event_ids = []
        start = time.time()
        while time.time() < (start + timeout):

            events = self.conn.describe_stack_events(stack_name)
            events.reverse()

            for event in events:
                if event.event_id not in seen_event_ids:
                    seen_event_ids.append(event.event_id)
                    if event.timestamp > valid_from_timestamp:

                        if event.resource_type == "AWS::CloudFormation::Stack":
                            self.logger.debug(event)

                            if event.resource_status == expected_event:
                                return event
                            if event.resource_status.endswith("_FAILED"):
                                raise CfnStackActionFailedException(
                                    "Stack is in {0} state".format(event.resource_status))
                            if event.resource_status.endswith("ROLLBACK_IN_PROGRESS"):
                                self.logger.error(
                                    "Failed to create stack (Reason: {1})".format(event.logical_resource_id,
                                                                                  event.resource_status_reason))
                            if event.resource_status.endswith("ROLLBACK_COMPLETE"):
                                raise CfnStackActionFailedException("Rollback occured")
                        else:
                            if event.resource_status.endswith("_FAILED"):
                                self.logger.error("Failed to create {0} (Reason: {1})".format(event.logical_resource_id,
                                                                                              event.resource_status_reason))
                            else:
                                self.logger.info(event)

            time.sleep(10)
        raise CfnStackActionFailedException(
            "Timeout occurred waiting for '{0}' on stack {1}".format(expected_event, stack_name))

    def wait_for_stack_action_to_complete(self, stack_name, action, timeout):

        allowed_actions = ["create", "update", "delete"]
        assert action.lower() in allowed_actions, "action argument must be one of {0}".format(allowed_actions)

        time_jitter_window = timedelta(seconds=10)
        minimum_event_timestamp = get_cfn_api_server_time() - time_jitter_window
        expected_start_event = action.upper() + "_IN_PROGRESS"

        start_event = self.wait_for_stack_events(stack_name,
                                                 expected_start_event,
                                                 minimum_event_timestamp,
                                                 timeout=120)

        self.logger.info("Stack {0} started".format(action))

        minimum_event_timestamp = start_event.timestamp
        expected_complete_event = action.upper() + "_COMPLETE"

        end_event = self.wait_for_stack_events(stack_name,
                                               expected_complete_event,
                                               minimum_event_timestamp,
                                               timeout)

        elapsed = end_event.timestamp - start_event.timestamp
        self.logger.info("Stack {0} completed after {1}s".format(action, elapsed.seconds))


if __name__ == "__main__":
    cfn = CloudFormation()
    print(cfn.get_stacks_dict())



from boto import ec2
from cfn_sphere.exceptions import CfnSphereException


class Ec2Api(object):
    def __init__(self, region="eu-west-1"):
        self.conn = ec2.connect_to_region(region)

    def get_taupage_images(self):
        filters = {'name': ['Taupage-AMI-*'],
                   'is-public': ['false'],
                   'state': ['available'],
                   'root-device-type': ['ebs']
                   }

        response = self.conn.get_all_images(executable_by=["self"], filters=filters)

        if not response:
            raise CfnSphereException("Could not find any private and available Taupage AMI")

        return response

    def get_latest_taupage_image_id(self):
        images = {image.creationDate: image.id for image in self.get_taupage_images()}

        creation_dates = images.keys()
        creation_dates.sort(reverse=True)
        return images[creation_dates[0]]



from boto import connect_s3
from urlparse import urlparse


class S3(object):
    def __init__(self):
        self.conn = connect_s3()

    @staticmethod
    def _parse_url(url):
        url_components = urlparse(url)
        protocol = url_components.scheme
        bucket_name = url_components.netloc
        key = url_components.path.strip('/')
        return protocol, bucket_name, key

    def get_contents_from_url(self, url):
        (protocol, bucket_name, key_name) = self._parse_url(url)
        bucket = self.conn.get_bucket(bucket_name)
        key = bucket.get_key(key_name)
        return key.get_contents_as_string()



import logging
import json
import yaml
import urllib2
import datetime
from cfn_sphere.exceptions import CfnSphereException


def get_logger(root=False):
    logging.basicConfig(format='%(asctime)s %(levelname)s %(module)s: %(message)s',
                        datefmt='%d.%m.%Y %H:%M:%S')
    if root:
        return logging.getLogger('cfn_sphere')
    else:
        return logging.getLogger('cfn_sphere.{0}'.format(__name__))


def convert_file(file_path):
    if file_path.lower().endswith('.json'):
        convert = convert_json_to_yaml_string
    elif file_path.lower().endswith('.yml'):
        convert = convert_yaml_to_json_string
    elif file_path.lower().endswith('.yaml'):
        convert = convert_yaml_to_json_string
    else:
        raise Exception('Unknown file extension. Please use .yaml, .yml or .json!')

    with open(file_path, 'r') as filestream:
        return convert(filestream.read())


def get_pretty_parameters_string(parameter_dict):
    parameters_string = ""
    for key, value in parameter_dict.items():
        parameters_string = parameters_string + "{0} = {1}\n".format(key, value)

    return parameters_string


def convert_json_to_yaml_string(data):
    if not data:
        return ''
    return yaml.safe_dump(json.loads(data), default_flow_style=False)


def convert_yaml_to_json_string(data):
    if not data:
        return '{}'
    return json.dumps(yaml.load(data), indent=2)


def convert_dict_to_json_string(data):
    return json.dumps(data, indent=2)


def get_message_from_boto_server_error(boto_server_error):
    return boto_server_error.message


def get_cfn_api_server_time():
    url = "http://aws.amazon.com"

    try:
        header_date = urllib2.urlopen(url).info().get('Date')
        return datetime.datetime.strptime(header_date, '%a, %d %b %Y %H:%M:%S GMT')
    except Exception as e:
        raise CfnSphereException("Could not get AWS server time from {0}. Error: {1}".format(url, e))


def get_latest_version():
    try:
        package_info = get_pypi_package_description()
        return package_info["info"]["version"]
    except Exception:
        return None


def get_pypi_package_description():
    url = "https://pypi.python.org/pypi/cfn-sphere/json"

    response = urllib2.urlopen(url, timeout=2)
    return json.load(response)



import json
from cfn_sphere.util import get_logger


class CloudFormationTemplate(object):
    def __init__(self, body_dict, name):
        self.logger = get_logger()
        self.name = name
        self.body_dict = body_dict
        self.template_format_version = body_dict.get('AWSTemplateFormatVersion', "2010-09-09")
        self.description = body_dict.get('Description', "")
        self.parameters = body_dict.get('Parameters', {})
        self.resources = body_dict.get('Resources', {})
        self.outputs = body_dict.get('Outputs', {})
        self.post_custom_resources = body_dict.get('PostCustomResources', {})

    def get_template_body_dict(self):
        return {
            'AWSTemplateFormatVersion': self.template_format_version,
            'Description': self.description,
            'Parameters': self.parameters,
            'Resources': self.resources,
            'Outputs': self.outputs
        }

    def get_template_json(self):
        return json.dumps(self.get_template_body_dict(), indent=2)
